Remove debug yield statements
[urcu.git] / urcu.c
diff --git a/urcu.c b/urcu.c
index 232a57bb718fbf12be651cebd6a808b01976c924..018e09b8601b041621fc6f72afdaeb57839e895c 100644 (file)
--- a/urcu.c
+++ b/urcu.c
 
 pthread_mutex_t urcu_mutex = PTHREAD_MUTEX_INITIALIZER;
 
-/* Global quiescent period parity */
-int urcu_qparity;
+/*
+ * Global grace period counter.
+ * Contains the current RCU_GP_CTR_BIT.
+ * Also has a RCU_GP_CTR_BIT of 1, to accelerate the reader fast path.
+ */
+long urcu_gp_ctr = RCU_GP_COUNT;
 
-int __thread urcu_active_readers[2];
+long __thread urcu_active_readers;
 
 /* Thread IDs of registered readers */
 #define INIT_NUM_THREADS 4
 
 struct reader_data {
        pthread_t tid;
-       int *urcu_active_readers;
+       long *urcu_active_readers;
 };
 
+#ifdef DEBUG_YIELD
+unsigned int yield_active;
+unsigned int __thread rand_yield;
+#endif
+
 static struct reader_data *reader_data;
 static int num_readers, alloc_readers;
+#ifndef DEBUG_FULL_MB
 static int sig_done;
+#endif
 
-void rcu_write_lock(void)
+void internal_urcu_lock(void)
 {
        int ret;
        ret = pthread_mutex_lock(&urcu_mutex);
@@ -46,7 +57,7 @@ void rcu_write_lock(void)
        }
 }
 
-void rcu_write_unlock(void)
+void internal_urcu_unlock(void)
 {
        int ret;
 
@@ -60,24 +71,28 @@ void rcu_write_unlock(void)
 /*
  * called with urcu_mutex held.
  */
-static int switch_next_urcu_qparity(void)
+static void switch_next_urcu_qparity(void)
 {
-       int old_parity = urcu_qparity;
-       urcu_qparity = 1 - old_parity;
-       return old_parity;
+       urcu_gp_ctr ^= RCU_GP_CTR_BIT;
 }
 
+#ifdef DEBUG_FULL_MB
+static void force_mb_all_threads(void)
+{
+       smp_mb();
+}
+#else
 static void force_mb_all_threads(void)
 {
        struct reader_data *index;
        /*
* Ask for each threads to execute a mb() so we can consider the
         * compiler barriers around rcu read lock as real memory barriers.
         */
        if (!reader_data)
                return;
        sig_done = 0;
-       mb();   /* write sig_done before sending the signals */
+       smp_mb();       /* write sig_done before sending the signals */
        for (index = reader_data; index < reader_data + num_readers; index++)
                pthread_kill(index->tid, SIGURCU);
        /*
@@ -86,10 +101,11 @@ static void force_mb_all_threads(void)
         */
        while (sig_done < num_readers)
                barrier();
-       mb();   /* read sig_done before ending the barrier */
+       smp_mb();       /* read sig_done before ending the barrier */
 }
+#endif
 
-void wait_for_quiescent_state(int parity)
+void wait_for_quiescent_state(void)
 {
        struct reader_data *index;
 
@@ -101,53 +117,65 @@ void wait_for_quiescent_state(int parity)
                /*
                 * BUSY-LOOP.
                 */
-               while (index->urcu_active_readers[parity] != 0)
+               while (rcu_old_gp_ongoing(index->urcu_active_readers))
                        barrier();
        }
-       /*
-        * Locally : read *index->urcu_active_readers before freeing old
-        * pointer.
-        * Remote (reader threads) : Order urcu_qparity update and other
-        * thread's quiescent state counter read.
-        */
-       force_mb_all_threads();
 }
 
-/*
- * Return old pointer, OK to free, no more reference exist.
- * Called under rcu_write_lock.
- */
-void *urcu_publish_content(void **ptr, void *new)
+void synchronize_rcu(void)
 {
-       int ret, prev_parity;
-       void *oldptr;
+       /* All threads should read qparity before accessing data structure
+        * where new ptr points to. */
+       /* Write new ptr before changing the qparity */
+       force_mb_all_threads();
+
+       internal_urcu_lock();
+
+       switch_next_urcu_qparity();     /* 0 -> 1 */
 
        /*
-        * We can publish the new pointer before we change the current qparity.
-        * Readers seeing the new pointer while being in the previous qparity
-        * window will make us wait until the end of the quiescent state before
-        * we release the unrelated memory area. However, given we hold the
-        * urcu_mutex, we are making sure that no further garbage collection can
-        * occur until we release the mutex, therefore we guarantee that this
-        * given reader will have completed its execution using the new pointer
-        * when the next quiescent state window will be over.
+        * Must commit qparity update to memory before waiting for parity
+        * 0 quiescent state. Failure to do so could result in the writer
+        * waiting forever while new readers are always accessing data (no
+        * progress).
         */
-       oldptr = *ptr;
-       *ptr = new;
-       /* All threads should read qparity before ptr */
-       /* Write ptr before changing the qparity */
-       force_mb_all_threads();
-       prev_parity = switch_next_urcu_qparity();
+       smp_mb();
 
        /*
         * Wait for previous parity to be empty of readers.
         */
-       wait_for_quiescent_state(prev_parity);
+       wait_for_quiescent_state();     /* Wait readers in parity 0 */
+
        /*
-        * Deleting old data is ok !
+        * Must finish waiting for quiescent state for parity 0 before
+        * committing qparity update to memory. Failure to do so could result in
+        * the writer waiting forever while new readers are always accessing
+        * data (no progress).
         */
-       
-       return oldptr;
+       smp_mb();
+
+       switch_next_urcu_qparity();     /* 1 -> 0 */
+
+       /*
+        * Must commit qparity update to memory before waiting for parity
+        * 1 quiescent state. Failure to do so could result in the writer
+        * waiting forever while new readers are always accessing data (no
+        * progress).
+        */
+       smp_mb();
+
+       /*
+        * Wait for previous parity to be empty of readers.
+        */
+       wait_for_quiescent_state();     /* Wait readers in parity 1 */
+
+       internal_urcu_unlock();
+
+       /* All threads should finish using the data referred to by old ptr
+        * before decrementing their urcu_active_readers count */
+       /* Finish waiting for reader threads before letting the old ptr being
+        * freed. */
+       force_mb_all_threads();
 }
 
 void urcu_add_reader(pthread_t id)
@@ -171,7 +199,7 @@ void urcu_add_reader(pthread_t id)
        }
        reader_data[num_readers].tid = id;
        /* reference to the TLS of _this_ reader thread. */
-       reader_data[num_readers].urcu_active_readers = urcu_active_readers;
+       reader_data[num_readers].urcu_active_readers = &urcu_active_readers;
        num_readers++;
 }
 
@@ -185,7 +213,7 @@ void urcu_remove_reader(pthread_t id)
 
        assert(reader_data != NULL);
        for (index = reader_data; index < reader_data + num_readers; index++) {
-               if (index->tid == id) {
+               if (pthread_equal(index->tid, id)) {
                        memcpy(index, &reader_data[num_readers - 1],
                                sizeof(struct reader_data));
                        reader_data[num_readers - 1].tid = 0;
@@ -200,22 +228,22 @@ void urcu_remove_reader(pthread_t id)
 
 void urcu_register_thread(void)
 {
-       rcu_write_lock();
+       internal_urcu_lock();
        urcu_add_reader(pthread_self());
-       rcu_write_unlock();
+       internal_urcu_unlock();
 }
 
 void urcu_unregister_thread(void)
 {
-       pthread_t self = pthread_self();
-       rcu_write_lock();
+       internal_urcu_lock();
        urcu_remove_reader(pthread_self());
-       rcu_write_unlock();
+       internal_urcu_unlock();
 }
 
+#ifndef DEBUG_FULL_MB
 void sigurcu_handler(int signo, siginfo_t *siginfo, void *context)
 {
-       mb();
+       smp_mb();
        atomic_inc(&sig_done);
 }
 
@@ -245,3 +273,4 @@ void __attribute__((destructor)) urcu_exit(void)
        assert(act.sa_sigaction == sigurcu_handler);
        free(reader_data);
 }
+#endif
This page took 0.025299 seconds and 4 git commands to generate.
@@ -60,24 +71,28 @@ void rcu_write_unlock(void)
        struct reader_data *index;
        /*
-        * Ask for each threads to execute a mb() so we can consider the
+        * Ask for each threads to execute a smp_mb() so we can consider the
         * compiler barriers around rcu read lock as real memory barriers.
         */